@@ -8,7 +8,7 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "history/view/media/history_view_media_grouped.h"
 
 #include "history/history_item_components.h"
-#include "history/history_message.h"
+#include "history/history_item.h"
 #include "history/history.h"
 #include "history/view/history_view_element.h"
 #include "history/view/history_view_cursor_state.h"
@@ -22,6 +22,7 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "ui/chat/message_bubble.h"
 #include "ui/text/text_options.h"
 #include "ui/painter.h"
+#include "ui/power_saving.h"
 #include "layout/layout_selection.h"
 #include "styles/style_chat.h"
 
@@ -117,6 +118,7 @@ QSize GroupedMedia::countOptimalSize() {
 	if (_mode == Mode::Column) {
 		for (const auto &part : _parts) {
 			const auto &media = part.content;
+			media->setBubbleRounding(bubbleRounding());
 			media->initDimensions();
 			accumulate_max(maxWidth, media->maxWidth());
 		}
@@ -244,15 +246,23 @@ void GroupedMedia::refreshParentId(
 	}
 }
 
-RectParts GroupedMedia::cornersFromSides(RectParts sides) const {
+Ui::BubbleRounding GroupedMedia::applyRoundingSides(
+		Ui::BubbleRounding already,
+		RectParts sides) const {
 	auto result = Ui::GetCornersFromSides(sides);
-	if (!isBubbleTop()) {
-		result &= ~(RectPart::TopLeft | RectPart::TopRight);
+	if (!(result & RectPart::TopLeft)) {
+		already.topLeft = Ui::BubbleCornerRounding::None;
 	}
-	if (!isRoundedInBubbleBottom() || !_caption.isEmpty()) {
-		result &= ~(RectPart::BottomLeft | RectPart::BottomRight);
+	if (!(result & RectPart::TopRight)) {
+		already.topRight = Ui::BubbleCornerRounding::None;
 	}
-	return result;
+	if (!(result & RectPart::BottomLeft)) {
+		already.bottomLeft = Ui::BubbleCornerRounding::None;
+	}
+	if (!(result & RectPart::BottomRight)) {
+		already.bottomRight = Ui::BubbleCornerRounding::None;
+	}
+	return already;
 }
 
 QMargins GroupedMedia::groupedPadding() const {
@@ -301,6 +311,11 @@ void GroupedMedia::draw(Painter &p, const PaintContext &context) const {
 	const auto textSelection = (_mode == Mode::Column)
 		&& !fullSelection
 		&& !IsSubGroupSelection(selection);
+	const auto inWebPage = (_parent->media() != this);
+	constexpr auto kSmall = Ui::BubbleCornerRounding::Small;
+	const auto rounding = inWebPage
+		? Ui::BubbleRounding{ kSmall, kSmall, kSmall, kSmall }
+		: adjustedBubbleRoundingWithCaption(_caption);
 	for (auto i = 0, count = int(_parts.size()); i != count; ++i) {
 		const auto &part = _parts[i];
 		const auto partContext = context.withSelection(fullSelection
@@ -324,7 +339,7 @@ void GroupedMedia::draw(Painter &p, const PaintContext &context) const {
 			partContext,
 			part.geometry.translated(0, groupPadding.top()),
 			part.sides,
-			cornersFromSides(part.sides),
+			applyRoundingSides(rounding, part.sides),
 			highlightOpacity,
 			&part.cacheKey,
 			&part.cache);
@@ -354,7 +369,8 @@ void GroupedMedia::draw(Painter &p, const PaintContext &context) const {
 			.palette = &stm->textPalette,
 			.spoiler = Ui::Text::DefaultSpoilerCache(),
 			.now = context.now,
-			.paused = context.paused,
+			.pausedEmoji = context.paused || On(PowerSaving::kEmojiChat),
+			.pausedSpoiler = context.paused || On(PowerSaving::kChatSpoiler),
 			.selection = context.selection,
 		});
 	} else if (_parent->media() == this) {
@@ -439,14 +455,16 @@ TextState GroupedMedia::textState(QPoint point, StateRequest request) const {
 			point,
 			InfoDisplayType::Image);
 		if (bottomInfoResult.link
-			|| bottomInfoResult.cursor != CursorState::None) {
+			|| bottomInfoResult.cursor != CursorState::None
+			|| bottomInfoResult.customTooltip) {
 			return bottomInfoResult;
 		}
 		if (const auto size = _parent->hasBubble() ? std::nullopt : _parent->rightActionSize()) {
 			auto fastShareLeft = (fullRight + st::historyFastShareLeft);
 			auto fastShareTop = (fullBottom - st::historyFastShareBottom - size->height());
 			if (QRect(fastShareLeft, fastShareTop, size->width(), size->height()).contains(point)) {
-				result.link = _parent->rightActionLink();
+				result.link = _parent->rightActionLink(point
+					- QPoint(fastShareLeft, fastShareTop));
 			}
 		}
 	}
@@ -532,7 +550,7 @@ TextForMimeData GroupedMedia::selectedText(
 			if (result.empty()) {
 				result = std::move(text);
 			} else {
-				result.append(qstr("\n\n")).append(std::move(text));
+				result.append(u"\n\n"_q).append(std::move(text));
 			}
 		}
 		selection = part.content->skipSelection(selection);
@@ -678,6 +696,13 @@ TextWithEntities GroupedMedia::getCaption() const {
 	return main()->getCaption();
 }
 
+void GroupedMedia::hideSpoilers() {
+	_caption.setSpoilerRevealed(false, anim::type::instant);
+	for (const auto &part : _parts) {
+		part.content->hideSpoilers();
+	}
+}
+
 Storage::SharedMediaTypesMask GroupedMedia::sharedMediaTypes() const {
 	return main()->sharedMediaTypes();
 }
@@ -736,7 +761,10 @@ void GroupedMedia::unloadHeavyPart() {
 }
 
 void GroupedMedia::parentTextUpdated() {
-	history()->owner().requestViewResize(_parent);
+	if (_parent->media() == this) {
+		refreshCaption();
+		history()->owner().requestViewResize(_parent);
+	}
 }
 
 bool GroupedMedia::needsBubble() const {
@@ -760,6 +788,7 @@ bool GroupedMedia::computeNeedBubble() const {
 			|| _parent->displayedReply()
 			|| _parent->displayForwardedFrom()
 			|| _parent->displayFromName()
+			|| _parent->displayedTopicButton()
 			) {
 			return true;
 		}
