@@ -7,7 +7,9 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 */
 #pragma once
 
+#include "base/timer.h"
 #include "ui/rp_widget.h"
+#include "ui/gl/gl_surface.h"
 #include "ui/widgets/dropdown_menu.h"
 #include "ui/effects/animations.h"
 #include "ui/effects/radial_animation.h"
@@ -16,6 +18,9 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "data/data_web_page.h"
 #include "data/data_cloud_themes.h" // Data::CloudTheme.
 #include "media/view/media_view_playback_controls.h"
+#include "media/view/media_view_open_common.h"
+
+class History;
 
 namespace Data {
 class PhotoMedia;
@@ -26,8 +31,19 @@ namespace Ui {
 class PopupMenu;
 class LinkButton;
 class RoundButton;
+class RpWindow;
 } // namespace Ui
 
+namespace Ui::GL {
+class Window;
+struct ChosenRenderer;
+enum class Backend;
+} // namespace Ui::GL
+
+namespace Platform {
+class OverlayWidgetHelper;
+} // namespace Platform
+
 namespace Window {
 namespace Theme {
 struct Preview;
@@ -41,35 +57,22 @@ struct TrackState;
 namespace Streaming {
 struct Information;
 struct Update;
+struct FrameWithInfo;
 enum class Error;
 } // namespace Streaming
 } // namespace Media
 
-namespace Media {
-namespace View {
+namespace Media::View {
 
 class GroupThumbs;
 class Pip;
 
-#if defined Q_OS_MAC && !defined OS_MAC_OLD
-#define USE_OPENGL_OVERLAY_WIDGET
-#endif // Q_OS_MAC && !OS_MAC_OLD
-
-#ifdef USE_OPENGL_OVERLAY_WIDGET
-using OverlayParent = Ui::RpWidgetWrap<QOpenGLWidget>;
-#else // USE_OPENGL_OVERLAY_WIDGET
-using OverlayParent = Ui::RpWidget;
-#endif // USE_OPENGL_OVERLAY_WIDGET
-
 class OverlayWidget final
-	: public OverlayParent
-	, private base::Subscriber
-	, public ClickHandlerHost
+	: public ClickHandlerHost
 	, private PlaybackControls::Delegate {
-	Q_OBJECT
-
 public:
 	OverlayWidget();
+	~OverlayWidget();
 
 	enum class TouchBarItemType {
 		Photo,
@@ -77,72 +80,56 @@ public:
 		None,
 	};
 
-	void showPhoto(not_null<PhotoData*> photo, HistoryItem *context);
-	void showPhoto(not_null<PhotoData*> photo, not_null<PeerData*> context);
-	void showDocument(
-		not_null<DocumentData*> document,
-		HistoryItem *context);
-	void showTheme(
-		not_null<DocumentData*> document,
-		const Data::CloudTheme &cloud);
+	[[nodiscard]] bool isActive() const;
+	[[nodiscard]] bool isHidden() const;
+	[[nodiscard]] bool isMinimized() const;
+	[[nodiscard]] bool isFullScreen() const;
+	[[nodiscard]] not_null<QWidget*> widget() const;
+	void hide();
+	void setCursor(style::cursor cursor);
+	void setFocus();
+	[[nodiscard]] bool takeFocusFrom(not_null<QWidget*> window) const;
+	void activate();
 
-	void leaveToChildEvent(QEvent *e, QWidget *child) override { // e -- from enterEvent() of child TWidget
-		updateOverState(OverNone);
-	}
-	void enterFromChildEvent(QEvent *e, QWidget *child) override { // e -- from leaveEvent() of child TWidget
-		updateOver(mapFromGlobal(QCursor::pos()));
-	}
+	void show(OpenRequest request);
 
-	void close();
+	//void leaveToChildEvent(QEvent *e, QWidget *child) override {
+	//	// e -- from enterEvent() of child TWidget
+	//	updateOverState(OverNone);
+	//}
+	//void enterFromChildEvent(QEvent *e, QWidget *child) override {
+	//	// e -- from leaveEvent() of child TWidget
+	//	updateOver(mapFromGlobal(QCursor::pos()));
+	//}
 
 	void activateControls();
-	void onDocClick();
-
-	PeerData *ui_getPeerForMouseAction();
+	void close();
+	void minimize();
+	void toggleFullScreen();
+	void toggleFullScreen(bool fullscreen);
 
 	void notifyFileDialogShown(bool shown);
 
 	void clearSession();
 
-	~OverlayWidget();
-
 	// ClickHandlerHost interface
 	void clickHandlerActiveChanged(const ClickHandlerPtr &p, bool active) override;
 	void clickHandlerPressedChanged(const ClickHandlerPtr &p, bool pressed) override;
 
-private slots:
-	void onHideControls(bool force = false);
-
-	void onScreenResized(int screen);
-
-	void onToMessage();
-	void onSaveAs();
-	void onDownload();
-	void onSaveCancel();
-	void onShowInFolder();
-	void onForward();
-	void onDelete();
-	void onOverview();
-	void onCopy();
-	void onMenuDestroy(QObject *obj);
-	void receiveMouse();
-	void onAttachedStickers();
-
-	void onDropdown();
-
-	void onTouchTimer();
-
-	void updateImage();
+	rpl::lifetime &lifetime();
 
 private:
 	struct Streamed;
 	struct PipWrap;
+	class Renderer;
+	class RendererSW;
+	class RendererGL;
 
+	// If changing, see paintControls()!
 	enum OverState {
 		OverNone,
 		OverLeftNav,
 		OverRightNav,
-		OverClose,
 		OverHeader,
 		OverName,
 		OverDate,
@@ -153,28 +140,60 @@ private:
 		OverVideo,
 	};
 	struct Entity {
-		base::optional_variant<
+		std::variant<
+			v::null_t,
 			not_null<PhotoData*>,
 			not_null<DocumentData*>> data;
-		HistoryItem *item;
+		HistoryItem *item = nullptr;
+		MsgId topicRootId = 0;
+	};
+	enum class SavePhotoVideo {
+		None,
+		QuickSave,
+		SaveAs,
+	};
+	struct ContentGeometry {
+		QRectF rect;
+		qreal rotation = 0.;
+		qreal controlsOpacity = 0.;
+	};
+	struct StartStreaming {
+		StartStreaming() : continueStreaming(false), startTime(0) {
+		}
+		StartStreaming(bool continueStreaming, crl::time startTime)
+		: continueStreaming(continueStreaming)
+		, startTime(startTime) {
+		}
+		const bool continueStreaming = false;
+		const crl::time startTime = 0;
 	};
 
-	void paintEvent(QPaintEvent *e) override;
+	[[nodiscard]] not_null<QWindow*> window() const;
+	[[nodiscard]] int width() const;
+	[[nodiscard]] int height() const;
+	void update();
+	void update(const QRegion &region);
 
-	void keyPressEvent(QKeyEvent *e) override;
-	void wheelEvent(QWheelEvent *e) override;
-	void mousePressEvent(QMouseEvent *e) override;
-	void mouseDoubleClickEvent(QMouseEvent *e) override;
-	void mouseMoveEvent(QMouseEvent *e) override;
-	void mouseReleaseEvent(QMouseEvent *e) override;
-	void contextMenuEvent(QContextMenuEvent *e) override;
-	void touchEvent(QTouchEvent *e);
+	[[nodiscard]] Ui::GL::ChosenRenderer chooseRenderer(
+		Ui::GL::Backend backend);
+	void paint(not_null<Renderer*> renderer);
 
-	bool eventHook(QEvent *e) override;
-	bool eventFilter(QObject *obj, QEvent *e) override;
-
-	void setVisibleHook(bool visible) override;
+	void setupWindow();
+	void orderWidgets();
+	void showAndActivate();
+	void handleMousePress(QPoint position, Qt::MouseButton button);
+	void handleMouseRelease(QPoint position, Qt::MouseButton button);
+	void handleMouseMove(QPoint position);
+	bool handleContextMenu(std::optional<QPoint> position);
+	bool handleDoubleClick(QPoint position, Qt::MouseButton button);
+	bool handleTouchEvent(not_null<QTouchEvent*> e);
+	void handleWheelEvent(not_null<QWheelEvent*> e);
+	void handleKeyPress(not_null<QKeyEvent*> e);
 
+	void toggleApplicationEventFilter(bool install);
+	bool filterApplicationEvent(
+		not_null<QObject*> object,
+		not_null<QEvent*> e);
 	void setSession(not_null<Main::Session*> session);
 
 	void playbackControlsPlay() override;
@@ -186,7 +205,7 @@ private:
 	void playbackControlsVolumeToggled() override;
 	void playbackControlsVolumeChangeFinished() override;
 	void playbackControlsSpeedChanged(float64 speed) override;
-	float64 playbackControlsCurrentSpeed() override;
+	float64 playbackControlsCurrentSpeed(bool lastNonDefault) override;
 	void playbackControlsToFullScreen() override;
 	void playbackControlsFromFullScreen() override;
 	void playbackControlsToPictureInPicture() override;
@@ -198,27 +217,64 @@ private:
 	void playbackPauseMusic();
 	void switchToPip();
 
+	void hideControls(bool force = false);
+	void subscribeToScreenGeometry();
+
+	void toMessage();
+	void saveAs();
+	void downloadMedia();
+	void saveCancel();
+	void showInFolder();
+	void forwardMedia();
+	void deleteMedia();
+	void showMediaOverview();
+	void copyMedia();
+	void receiveMouse();
+	void showAttachedStickers();
+	void showDropdown();
+	void handleTouchTimer();
+	void handleDocumentClick();
+	void updateImage();
+
+	void clearBeforeHide();
+	void clearAfterHide();
+
 	void assignMediaPointer(DocumentData *document);
 	void assignMediaPointer(not_null<PhotoData*> photo);
 
 	void updateOver(QPoint mpos);
-	void moveToScreen(bool force = false);
+	void initFullScreen();
+	void initNormalGeometry();
+	void savePosition();
+	void moveToScreen(bool inMove = false);
+	void updateGeometry(bool inMove = false);
+	void updateGeometryToScreen(bool inMove = false);
 	bool moveToNext(int delta);
 	void preloadData(int delta);
 
+	void handleScreenChanged(QScreen *screen);
+
+	bool contentCanBeSaved() const;
+	void checkForSaveLoaded();
+
 	Entity entityForUserPhotos(int index) const;
 	Entity entityForSharedMedia(int index) const;
 	Entity entityForCollage(int index) const;
 	Entity entityByIndex(int index) const;
 	Entity entityForItemId(const FullMsgId &itemId) const;
 	bool moveToEntity(const Entity &entity, int preloadDelta = 0);
-	void setContext(base::optional_variant<
-		not_null<HistoryItem*>,
+
+	struct ItemContext {
+		not_null<HistoryItem*> item;
+		MsgId topicRootId = 0;
+	};
+	void setContext(std::variant<
+		v::null_t,
+		ItemContext,
 		not_null<PeerData*>> context);
 
 	void refreshLang();
 	void showSaveMsgFile();
-	void updateMixerVideoVolume() const;
 
 	struct SharedMedia;
 	using SharedMediaType = SharedMediaWithLastSlice::Type;
@@ -246,8 +302,8 @@ private:
 	[[nodiscard]] Data::FileOrigin fileOrigin() const;
 	[[nodiscard]] Data::FileOrigin fileOrigin(const Entity& entity) const;
 
-	void refreshFromLabel(HistoryItem *item);
-	void refreshCaption(HistoryItem *item);
+	void refreshFromLabel();
+	void refreshCaption();
 	void refreshMediaViewer();
 	void refreshNavVisibility();
 	void refreshGroupThumbs();
@@ -255,21 +311,23 @@ private:
 	void dropdownHidden();
 	void updateDocSize();
 	void updateControls();
-	void updateActions();
+	void updateControlsGeometry();
+
+	using MenuCallback = Fn<void(
+		const QString &,
+		Fn<void()>,
+		const style::icon *)>;
+	void fillContextMenuActions(const MenuCallback &addAction);
+
 	void resizeCenteredControls();
 	void resizeContentByScreenSize();
+	void recountSkipTop();
 
-	void showDocument(
-		not_null<DocumentData*> document,
-		HistoryItem *context,
-		const Data::CloudTheme &cloud,
-		bool continueStreaming);
-	void displayPhoto(not_null<PhotoData*> photo, HistoryItem *item);
+	void displayPhoto(not_null<PhotoData*> photo);
 	void displayDocument(
 		DocumentData *document,
-		HistoryItem *item,
 		const Data::CloudTheme &cloud = Data::CloudTheme(),
-		bool continueStreaming = false);
+		const StartStreaming &startStreaming = StartStreaming());
 	void displayFinished();
 	void redisplayContent();
 	void findCurrent();
@@ -278,28 +336,34 @@ private:
 	void setZoomLevel(int newZoom, bool force = false);
 
 	void updatePlaybackState();
+	void seekRelativeTime(crl::time time);
+	void restartAtProgress(float64 progress);
 	void restartAtSeekPosition(crl::time position);
 
 	void refreshClipControllerGeometry();
 	void refreshCaptionGeometry();
 
-	bool initStreaming(bool continueStreaming = false);
-	void startStreamingPlayer();
+	bool initStreaming(
+		const StartStreaming &startStreaming = StartStreaming());
+	void startStreamingPlayer(const StartStreaming &startStreaming);
 	void initStreamingThumbnail();
 	void streamingReady(Streaming::Information &&info);
 	[[nodiscard]] bool createStreamingObjects();
 	void handleStreamingUpdate(Streaming::Update &&update);
 	void handleStreamingError(Streaming::Error &&error);
+	void updatePowerSaveBlocker(const Player::TrackState &state);
 
 	void initThemePreview();
 	void destroyThemePreview();
 	void updateThemePreviewGeometry();
 
-	void documentUpdated(DocumentData *doc);
-	void changingMsgId(not_null<HistoryItem*> row, MsgId oldId);
+	void documentUpdated(not_null<DocumentData*> document);
+	void changingMsgId(FullMsgId newId, MsgId oldId);
 
-	[[nodiscard]] int contentRotation() const;
-	[[nodiscard]] QRect contentRect() const;
+	[[nodiscard]] int finalContentRotation() const;
+	[[nodiscard]] QRect finalContentRect() const;
+	[[nodiscard]] ContentGeometry contentGeometry() const;
+	void updateContentRect();
 	void contentSizeChanged();
 
 	// Radial animation interface.
@@ -323,13 +387,42 @@ private:
 	void zoomReset();
 	void zoomUpdate(int32 &newZoom);
 
-	void paintRadialLoading(Painter &p, bool radial, float64 radialOpacity);
+	void paintRadialLoading(not_null<Renderer*> renderer);
 	void paintRadialLoadingContent(
 		Painter &p,
 		QRect inner,
 		bool radial,
 		float64 radialOpacity) const;
-	void paintThemePreview(Painter &p, QRect clip);
+	void paintThemePreviewContent(Painter &p, QRect outer, QRect clip);
+	void paintDocumentBubbleContent(
+		Painter &p,
+		QRect outer,
+		QRect icon,
+		QRect clip) const;
+	void paintSaveMsgContent(Painter &p, QRect outer, QRect clip);
+	void paintControls(not_null<Renderer*> renderer, float64 opacity);
+	void paintFooterContent(
+		Painter &p,
+		QRect outer,
+		QRect clip,
+		float64 opacity);
+	[[nodiscard]] QRect footerGeometry() const;
+	void paintCaptionContent(
+		Painter &p,
+		QRect outer,
+		QRect clip,
+		float64 opacity);
+	[[nodiscard]] QRect captionGeometry() const;
+	void paintGroupThumbsContent(
+		Painter &p,
+		QRect outer,
+		QRect clip,
+		float64 opacity);
+
+	[[nodiscard]] float64 controlOpacity(
+		float64 progress,
+		bool nonbright = false) const;
+	[[nodiscard]] bool isSaveMsgShown() const;
 
 	void updateOverRect(OverState state);
 	bool updateOverState(OverState newState);
@@ -341,25 +434,52 @@ private:
 	void validatePhotoImage(Image *image, bool blurred);
 	void validatePhotoCurrentImage();
 
+	[[nodiscard]] bool hasCopyMediaRestriction() const;
+	[[nodiscard]] bool showCopyMediaRestriction();
+
 	[[nodiscard]] QSize flipSizeByRotation(QSize size) const;
 
 	void applyVideoSize();
 	[[nodiscard]] bool videoShown() const;
 	[[nodiscard]] QSize videoSize() const;
 	[[nodiscard]] bool videoIsGifOrUserpic() const;
-	[[nodiscard]] QImage videoFrame() const;
-	[[nodiscard]] QImage videoFrameForDirectPaint() const;
-	[[nodiscard]] QImage transformVideoFrame(QImage frame) const;
-	[[nodiscard]] QImage transformStaticContent(QPixmap content) const;
+	[[nodiscard]] QImage videoFrame() const; // ARGB (changes prepare format)
+	[[nodiscard]] QImage currentVideoFrameImage() const; // RGB (may convert)
+	[[nodiscard]] Streaming::FrameWithInfo videoFrameWithInfo() const; // YUV
+	[[nodiscard]] int streamedIndex() const;
+	[[nodiscard]] QImage transformedShownContent() const;
+	[[nodiscard]] QImage transformShownContent(
+		QImage content,
+		int rotation) const;
 	[[nodiscard]] bool documentContentShown() const;
 	[[nodiscard]] bool documentBubbleShown() const;
-	void paintTransformedVideoFrame(Painter &p);
-	void paintTransformedStaticContent(Painter &p);
+	void setStaticContent(QImage image);
+	[[nodiscard]] bool contentShown() const;
+	[[nodiscard]] bool opaqueContentShown() const;
 	void clearStreaming(bool savePosition = true);
 	bool canInitStreaming() const;
 
-	QBrush _transparentBrush;
+	[[nodiscard]] bool topShadowOnTheRight() const;
+	void applyHideWindowWorkaround();
 
+	Window::SessionController *findWindow(bool switchTo = true) const;
+
+	bool _opengl = false;
+	const std::unique_ptr<Ui::GL::Window> _wrap;
+	const not_null<Ui::RpWindow*> _window;
+	const std::unique_ptr<Platform::OverlayWidgetHelper> _helper;
+	const not_null<Ui::RpWidget*> _body;
+	const std::unique_ptr<Ui::RpWidget> _titleBugWorkaround;
+	const std::unique_ptr<Ui::RpWidgetWrap> _surface;
+	const not_null<QWidget*> _widget;
+	QRect _normalGeometry;
+	bool _wasWindowedMode = false;
+	bool _fullscreenInited = false;
+	bool _normalGeometryInited = false;
+	bool _fullscreen = true;
+	bool _windowed = false;
+
+	base::weak_ptr<Window::Controller> _openedFrom;
 	Main::Session *_session = nullptr;
 	rpl::lifetime _sessionLifetime;
 	PhotoData *_photo = nullptr;
@@ -377,10 +497,12 @@ private:
 	std::unique_ptr<Collage> _collage;
 	std::optional<WebPageCollage> _collageData;
 
-	QRect _closeNav, _closeNavIcon;
-	QRect _leftNav, _leftNavIcon, _rightNav, _rightNavIcon;
+	QRect _leftNav, _leftNavOver, _leftNavIcon;
+	QRect _rightNav, _rightNavOver, _rightNavIcon;
 	QRect _headerNav, _nameNav, _dateNav;
-	QRect _rotateNav, _rotateNavIcon, _saveNav, _saveNavIcon, _moreNav, _moreNavIcon;
+	QRect _rotateNav, _rotateNavOver, _rotateNavIcon;
+	QRect _saveNav, _saveNavOver, _saveNavIcon;
+	QRect _moreNav, _moreNavOver, _moreNavIcon;
 	bool _leftNavVisible = false;
 	bool _rightNavVisible = false;
 	bool _saveVisible = false;
@@ -404,6 +526,8 @@ private:
 
 	int _width = 0;
 	int _height = 0;
+	int _skipTop = 0;
+	int _availableHeight = 0;
 	int _x = 0, _y = 0, _w = 0, _h = 0;
 	int _xStart = 0, _yStart = 0;
 	int _zoom = 0; // < 0 - out, 0 - none, > 0 - in
@@ -412,28 +536,42 @@ private:
 	QPoint _mStart;
 	bool _pressed = false;
 	int32 _dragging = 0;
-	QPixmap _staticContent;
+	QImage _staticContent;
+	bool _staticContentTransparent = false;
 	bool _blurred = true;
+	bool _reShow = false;
+
+	ContentGeometry _oldGeometry;
+	Ui::Animations::Simple _geometryAnimation;
+	rpl::lifetime _screenGeometryLifetime;
+	std::unique_ptr<QObject> _applicationEventFilter;
 
 	std::unique_ptr<Streamed> _streamed;
 	std::unique_ptr<PipWrap> _pip;
+	int _streamedCreated = 0;
+	bool _showAsPip = false;
 
 	const style::icon *_docIcon = nullptr;
 	style::color _docIconColor;
 	QString _docName, _docSize, _docExt;
 	int _docNameWidth = 0, _docSizeWidth = 0, _docExtWidth = 0;
 	QRect _docRect, _docIconRect;
+	QImage _docRectImage;
 	int _docThumbx = 0, _docThumby = 0, _docThumbw = 0;
 	object_ptr<Ui::LinkButton> _docDownload;
 	object_ptr<Ui::LinkButton> _docSaveAs;
 	object_ptr<Ui::LinkButton> _docCancel;
 
+	QRect _bottomShadowRect;
+	QRect _topShadowRect;
+	rpl::variable<bool> _topShadowRight = false;
+
 	QRect _photoRadialRect;
 	Ui::RadialAnimation _radial;
-	QImage _radialCache;
 
 	History *_migrated = nullptr;
 	History *_history = nullptr; // if conversation photos or files overview
+	MsgId _topicRootId = 0;
 	PeerData *_peer = nullptr;
 	UserData *_user = nullptr; // if user profile photos overview
 
@@ -450,9 +588,7 @@ private:
 	std::optional<int> _index; // Index in current _sharedMedia data.
 	std::optional<int> _fullIndex; // Index in full shared media.
 	std::optional<int> _fullCount;
-	FullMsgId _msgid;
-	bool _canForwardItem = false;
-	bool _canDeleteItem = false;
+	HistoryItem *_message = nullptr;
 
 	mtpRequestId _loadRequest = 0;
 
@@ -471,35 +607,30 @@ private:
 	};
 	ControlsState _controlsState = ControlsShown;
 	crl::time _controlsAnimStarted = 0;
-	QTimer _controlsHideTimer;
-	anim::value _controlsOpacity;
+	base::Timer _controlsHideTimer;
+	anim::value _controlsOpacity = { 1. };
 	bool _mousePressed = false;
 
-	Ui::PopupMenu *_menu = nullptr;
+	base::unique_qptr<Ui::PopupMenu> _menu;
 	object_ptr<Ui::DropdownMenu> _dropdown;
-	object_ptr<QTimer> _dropdownShowTimer;
-
-	struct ActionData {
-		QString text;
-		const char *member;
-	};
-	QList<ActionData> _actions;
+	base::Timer _dropdownShowTimer;
 
 	bool _receiveMouse = true;
 
 	bool _touchPress = false;
 	bool _touchMove = false;
 	bool _touchRightButton = false;
-	QTimer _touchTimer;
+	base::Timer _touchTimer;
 	QPoint _touchStart;
-	QPoint _accumScroll;
 
 	QString _saveMsgFilename;
-	crl::time _saveMsgStarted = 0;
-	anim::value _saveMsgOpacity;
 	QRect _saveMsg;
-	QTimer _saveMsgUpdater;
+	QImage _saveMsgImage;
 	Ui::Text::String _saveMsgText;
+	SavePhotoVideo _savePhotoVideoWhenLoaded = SavePhotoVideo::None;
+	// _saveMsgAnimation -> _saveMsgTimer -> _saveMsgAnimation.
+	Ui::Animations::Simple _saveMsgAnimation;
+	base::Timer _saveMsgTimer;
 
 	base::flat_map<OverState, crl::time> _animations;
 	base::flat_map<OverState, anim::value> _animationOpacities;
@@ -519,9 +650,8 @@ private:
 	object_ptr<Ui::RoundButton> _themeShare = { nullptr };
 	Data::CloudTheme _themeCloudData;
 
-	bool _wasRepainted = false;
+	std::unique_ptr<Ui::RpWidget> _hideWorkaround;
 
 };
 
-} // namespace View
-} // namespace Media
+} // namespace Media::View
